package ghmessen.messtasks;

import ghmessen.adapter.EAdapterException;

import java.util.ArrayList;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import ld.api.CASSYs;
import ld.api.ICASSY;
import ld.api.ITimer;
import ld.api.OscilloscopeStates;
import ld.api.TriggerBits;
import ld.api.ios.CASSYIO;
import ld.api.ios.CASSYIOs;
import ld.api.ios.CASSYQuantity;
import ld.api.ios.ICASSYQuantity;
import ld.api.ios.MeasurementMode;
import ld.api.ios.OscilloscopeError;

public class CassyVector extends ArrayList<CASSYs> {
	private static final long serialVersionUID = 1L;
	private static final boolean DEBUG = false;

	private CASSYIOs cassyios;
	private boolean quantitiesSelected;
	private ScheduledFuture<?> einzelMessung;
	private CountDownLatch einzelMessungComplete;
	private Boolean restartEinzelMessung;
	private OszilloskopMessung oszilloskopMessung;
	public CountDownLatch oszilloskopStarted;
	private CountDownLatch oszilloskopMessungComplete;
	private Boolean oszilloskopWerteVorhanden;
	private ScheduledThreadPoolExecutor timer;

	public CassyVector(ScheduledThreadPoolExecutor timer) {
		super();
		cassyios = new CASSYIOs();
		quantitiesSelected = false;
		cassyios.setAutoDefaults(true);
		einzelMessung = null;
		oszilloskopStarted = null;
		einzelMessungComplete = null;
		restartEinzelMessung = false;
		oszilloskopMessung = null;
		oszilloskopMessungComplete = null;
		oszilloskopWerteVorhanden = false;
		this.timer = timer;
		if (DEBUG)
			System.out.println("[" + this.getClass().toString()
					+ "] Konstruktor ausgeführt.");
	}

	@Override
	public boolean add(CASSYs c) {
		boolean result = super.add(c);
		for (CASSYIO io : c.getCASSYIOs())
			cassyios.add(io);
		starteMessung();
		return result;
	}

	public int getNr(ICASSYQuantity quantity) {
		ICASSY cassy = quantity.getInput().getCASSY();
		for (int index = 0; index < size(); index++)
			if (get(index) == cassy)
				return index;
		return -1;
	}

	public CASSYIOs getCASSYIOs() {
		return cassyios;
	}

	public boolean isQuantitiesSelected() {
		return quantitiesSelected;
	}
	
	public int getTimes() {
		return cassyios.getTimes();
	}

	public Double getTimes(int i) {
		return cassyios.getTimes(i);
	}

	@Override
	public boolean remove(Object o) {
		int index = indexOf(o);
		boolean result = super.remove(o);
		if (index < 0 || !(o instanceof CASSYs))
			return false;
		CASSYs c = (CASSYs) o;
		stoppeMessung();
		for (CASSYIO io : c.getCASSYIOs())
			cassyios.remove(io);
		if (size() >= 1) {
			starteMessung();
		}
		return result;
	}

	private boolean istOszilloskopMessung(double intervall) {
		// Falls beide Möglichkeiten erlaubt sind, entscheiden die Intervalle.
		boolean seriell = false;
		boolean einzelAllowed = cassyios.getSlowOscilloscopeAllowed();
		boolean osziAllowed = cassyios.getFastOscilloscopeAllowed();
		for(CASSYs c : this) {
			if(c.getCommPort().matches(".*tty.*"))
				seriell = true;
		}
		if(osziAllowed && (seriell && intervall < 0.1 || intervall < 0.01))
			return true;
		else if(einzelAllowed)
			return false;
		else if(osziAllowed)
			return true;
		else
			return false;
	}
	
	public void init(double intervall) {
		quantitiesSelected = false;
		for (CASSYIO io : cassyios) {
			if (io.size() > 0) {
				quantitiesSelected = true;
				break;
			}
		}
		if (istOszilloskopMessung(intervall))
			cassyios.setMeasurementMode(MeasurementMode.FastOscilloscope);
		else
			cassyios.setMeasurementMode(MeasurementMode.SlowOscilloscope);
	}

	//
	// Messungen
	//
	public boolean starteMessung() {
		if (DEBUG)
			System.out.println("[CassyVector] - starteMessung: "+System.currentTimeMillis());
		if (size() > 0 && timer != null) {
			cassyios.setMeasurementMode(MeasurementMode.SingleMeasurement);
			if (einzelMessung == null)
				einzelMessung = (ScheduledFuture<?>) timer
						.scheduleWithFixedDelay(new EinzelMessung(), 0, 100,
								TimeUnit.MILLISECONDS);
			return true;
		} else
			return false;
	}

	public boolean stoppeMessung() {
		if (DEBUG)
			System.out.println("[CassyVector] - stoppeMessung");
		if (timer != null && einzelMessung != null) {
			einzelMessung.cancel(false);
			if(einzelMessungComplete != null)
				try {
					einzelMessungComplete.await(100,TimeUnit.MILLISECONDS);
				} catch (InterruptedException e){ }
			einzelMessung = null;
		}
		return true;
	}

	public boolean verwendetTimer() {
		for (CASSYs cs : this)
			for (ICASSY c : cs)
				if (c instanceof ITimer)
					return true;
		return false;
	}

	public void setTriggerQuantity(String symbol){
		for(CASSYIO io : cassyios)
			for(CASSYQuantity q : io.getSelectedQuantities())
				if(q.getSymbol().equals(symbol)){
					cassyios.setTriggerQuantity(q);
					return;
				}
	}
	
	public void setTrigger(boolean ein) {
		cassyios.setTrigger(ein);
	}
	
	public void setTriggerFlanke(TriggerBits trigger){
		cassyios.setTriggerRising(!trigger.isAndEmpty(TriggerBits.Rising));
	}
	
	public void setTriggerValue(double value) {
		cassyios.setTriggerValue(value);
	}
	
	public void setOszilloskopeWiederholung(int wiederholung) {
		cassyios.setOscilloscopeCount(wiederholung);
	}

	public void setOszilloskopeIntervall(double interval) {
		cassyios.setOscilloscopeInterval(interval);
	}

	public void starteOszilloskope() throws EAdapterException {
		if (DEBUG)
			System.out.println("[CassyVector] - starteOszilloskope");
		if (size() > 0 && timer != null) {
			synchronized (oszilloskopWerteVorhanden) {
				oszilloskopWerteVorhanden = false;
			}
			synchronized (restartEinzelMessung) {
				restartEinzelMessung = einzelMessung != null;
				stoppeMessung();
			}
			if (oszilloskopMessung == null) {
				if(DEBUG)
					System.out.println("[CassyVector] - starteOszilloskope: Oszilloskopmessung (100ms delay).");
				oszilloskopStarted = new CountDownLatch(1);
				oszilloskopMessung = new OszilloskopMessung();
				timer.schedule(oszilloskopMessung, 0, TimeUnit.MILLISECONDS);
			}
		}
	}

	public String getOszilloskopError() {
		OscilloscopeError resultat = cassyios.getOscilloscopeError();
		if(resultat == OscilloscopeError.None)
			return null;
		if(DEBUG)
			System.err.println("[CassyVector] - getOszilloskopError("+resultat+")");
		return "Es ist der Oszilloskopfehler "+resultat.getValue()+" aufgetreten.";
	}
	
	public void stoppeOszilloskope() throws EAdapterException {
		if (DEBUG)
			System.out.println("[CassyVector] - stoppeOszilloskope");
		if(timer != null && oszilloskopMessung != null) {
			synchronized(oszilloskopMessung.doTerminate) {
				oszilloskopMessung.doTerminate = true;
			}
			if(oszilloskopMessungComplete != null) {
				try {
					oszilloskopMessungComplete.await(100, TimeUnit.MILLISECONDS);
				} catch (InterruptedException e) {}
			}
			oszilloskopMessung = null;
		}
		if (einzelMessung == null && restartEinzelMessung) {
			starteMessung();
			synchronized (restartEinzelMessung) {
				restartEinzelMessung = false;
			}
		}
	}

	public boolean istMesswertVorhanden() {
		if (einzelMessung != null)
			return true;
		return oszilloskopWerteVorhanden;
	}

	private class EinzelMessung implements Runnable {
		private static final boolean DEBUG = false;
		@Override
		public void run() {
			if (DEBUG)
				System.out.print("E");
			if(einzelMessungComplete != null && einzelMessungComplete.getCount() > 0)
				return;
			// es darf nur eine Messung zur Zeit laufen, abwarten ist nicht sinnvoll, andernfalls
			// können sich Messungen "aufstauen".
			einzelMessungComplete = new CountDownLatch(1);
			cassyios.doSingleMeasurement();
			einzelMessungComplete.countDown();
			// verhindert eine Unterbrechung der Einzelmessung, sondern stoppt nach einer Messung
			if (DEBUG)
				System.out.print(".");
		}
	}

	private class OszilloskopMessung implements Runnable {
		private static final boolean DEBUG = false;
		private OscilloscopeStates status = OscilloscopeStates.Stopped;
		public Boolean doTerminate = false;
		@Override
		public void run() {
			if(oszilloskopMessungComplete != null && oszilloskopMessungComplete.getCount() > 0) {
				System.err.println("[OszilloskopMessung] - run: Es läuft bereits eine Messung!");
				return;
			}
			oszilloskopMessungComplete = new CountDownLatch(1);
			while(!doTerminate && !cassyios.doOscilloscope()) {
				if(oszilloskopStarted != null && oszilloskopStarted.getCount() > 0)
					oszilloskopStarted.countDown();
				if (DEBUG)
					System.out.print("O");
				status = cassyios.getOscilloscopeStatus();
				synchronized(oszilloskopWerteVorhanden) {
					oszilloskopWerteVorhanden = status != OscilloscopeStates.TriggerWaiting;
				}
				if (DEBUG)
					System.out.print(".");
			}
			cassyios.stopOscilloscope();
			oszilloskopMessungComplete.countDown();
			if (DEBUG)
				System.out.println("O!");
		}
	}
}
